;;; Implementation of local
;;; ML Style local declaration. Can be used anywhere DEFINE can be
;;; used. Does not support the (DEFINE (name args ...) body ...)
;;; syntax.
;;; Usage:

;;; decl::= (DEFINE name exp)
;;;       | (BEGIN decl ...)
;;;       | (LOCAL (decl ...) decl decl ...)

(define *do-local  
  (let ((set-difference
	  (lambda (s1 s2)
	    (let loop ([s s1] [ans '()])
	      (cond
		[(null? s) (reverse ans)]
		[(member (car s) s2) (loop (cdr s) ans)]
		[else (loop (cdr s) (cons (car s) ans))])))))
    (lambda (x)
      (syntax-case x (define)
	[(((define local-name local-exp) ...)
	  (define name exp)
	  ...)
	 (with-syntax
	   ([(name* ...)
	     (set-difference (syntax (name ...)) (syntax (local-name ...)))]
	    [(ignored-name) (generate-temporaries '(ignored))])
	   (syntax
	     (begin
	       (define name* #f)
	       ...
	       (define ignored-name ; gensymed name
		 (internal-let ((define local-name local-exp) ...)
		   (set! name exp)
		   ...)))))]
	[(local-decls decl ...)
	 (with-syntax
	   ([local-defns (*flatten (syntax (local-decls ())))]
	    [(defn ...) (*flatten (syntax ((decl ...) ())))])
	   (*do-local (syntax (local-defns defn ...))))]))))

(define *flatten
  (lambda (x)
    (syntax-case x (define begin local)
      [(() defns) (syntax defns)]
      [(((define name exp) more ...) (defn ...))
       (*flatten (syntax ((more ...) (defn ... (define name exp)))))]
      [(((begin decl ...) more ...) defns)
       (*flatten (syntax ((decl ... more ...) defns)))]
      [(((local local-decls decl ...) more ...) defns)
       (with-syntax
	 ([decl (*do-local (syntax (local-decls decl ...)))])
	 (*flatten (syntax ((decl more ...) defns))))])))

(define-syntax internal-let
  (letrec ((remove-duplicates
	     (lambda (ls)
	       (if (null? ls) ls
		 (if (member (car ls) (cdr ls))
		   (remove-duplicates (cdr ls))
		   (cons (car ls) (remove-duplicates (cdr ls))))))))
    (lambda (x)
      (syntax-case x (define)
	[(_ ((define name exp) ...) body ...)
	 (with-syntax
	   ([(name* ...) (remove-duplicates (syntax (name ...)))])
	   (syntax
	     (let ([name* #f] ...)
	       (set! name exp) ...
	       body ...)))]
	[(_ decls body ...)
	 (with-syntax
	   ([defns (*flatten (syntax (decls ())))])
	   (syntax (internal-let defns body ...)))]))))

(define-syntax local
  (lambda (x)
    (syntax-case x ()
      [(_ more ...) (*do-local (syntax (more ...)))])))





